!**************************************************************************************************
! Fortran code to replicate Johri, Khan, and Sosa-Padilla's
! "Interest Rate Uncertainty and Sovereign Default Risk" (2022) -- 2nd round RR at the JIE.
! --------
! Some usage notes can be found thourghout the code but mostly in the companion README file.
!***************************************************************************************************

module param
DOUBLE PRECISION :: beta, sigma, b_inf, b_sup, y_inf, y_sup, eps_inf, eps_sup, std_eps,&
     rho, mean_y, std_y, pi_number, zero, width, prob_excl_end, mean_g,&
     r_prime, r_vol_inf, r_vol_sup, std_u_r, std_u_vol, rho_r,&
     rho_r_vol, mean_r, mean_r_vol, std_r_vol, d0, d0_tilde, d1, eta,&
     coupon, delta, kappa, kappa_rw, gov_spending

parameter (sigma = 2d+0, pi_number = 3.1415926535897932d+0, &
     zero = 0d+0,  width = 1.5d+0,  mean_g = 1.000d+0,&
     std_u_r = 1d+0, std_u_vol= 1d+0, rho_r = 0.908d+0,&
     rho_r_vol = 0.8742d+0, mean_r = 0d+0, mean_r_vol = -6.2869d+0, r_prime = 0.01d+0, eta = 0.2632d+0,&
     prob_excl_end = 0.0385d+0, std_eps = 0.027d+0, rho = 0.933d+0,  mean_y = -0.5d+0*std_eps**2,& !
     delta = 0.0742d+0, gov_spending = 0d+0)

! *********************************************************************************************************
! NOTE: TO SOLVE THE "BASIC" MODEL (WITHOUT FLUCTUATIONS IN r) YOU NEED TO SET THE FOLLOWING CHANGES:
! 1st: Set std_u_r = 0d+0 (this shuts down shocks to the level of r) in line 17 (above)
! 2nd: Set std_u_vol = 0d+0 (this shuts down shocks to the volatility of r) in line 17 (above)
! 3rd: Set r_num = 1 and r_vol =1 in line 42 (below). This speeds up the solution.
!
! NOTE: TO SOLCE THE "NEW SDF" MODEL IN APPENDIX B YOU NEED TO SET "kappa_rw" and "kappa" to the values
! reported in Appendix B. 
! *********************************************************************************************************
!y = log(income)
!sigma = coefficient of relative risk aversion
!r_prime = mean of the international interest rate
!rho = coefficient of autocorrelation in income
!std_eps = standard deviation of innovation to income shocks
!mean_y = unconditional mean of the growth trend shock
!std_y = standard deviation of growth trend shock

INTEGER :: b_num, y_num, eps_num, quad_num, nout, i_y_global, i_excl_global, r_num, r_vol_num, &
     i_b_global, i_default_global, indicator_2, i_b_next, i_b_zero, b_num_interm, counter

parameter (b_num= 200, y_num = 35, b_num_interm=60, r_num = 7, r_vol_num = 7)

!indicator_unix > 0, saves files in the same folder as .f90 file
!indicator_unix = 0, saves files in a new folder \graphs (which must exist prior to compiling .f90 file)

DOUBLE PRECISION :: b_grid(1:b_num), y_grid(1:y_num), default_grid(1:2), &
     std_r(1:r_vol_num),r_grid(r_num,r_vol_num),&
     r_vol_grid (1:r_vol_num), matrix_v_global(b_num, y_num, r_num, r_vol_num),&
     vector_v_global(y_num, r_num, r_vol_num),y_initial, b_initial, r_initial, r_vol_initial, &
     v1_matrix(y_num, r_num, r_vol_num), exp_v_vector(b_num),q_vector(b_num),&
     exp_v_excl_vector(b_num), b_global, y_global, r_global, r_vol_global, q_nodef(y_num,r_num, r_vol_num),&
     expected_SDF_matrix(y_num, r_num, r_vol_num), variance_r_level_epsilon_tomorrow(r_vol_num)


DOUBLE PRECISION, DIMENSION(b_num, y_num, r_num, r_vol_num) :: v_matrix
INTEGER, DIMENSION(b_num, y_num, r_num, r_vol_num) :: default_decision, b_next_matrix, b0_next_matrix_indice
DOUBLE PRECISION, DIMENSION(b_num, y_num, r_num, r_vol_num) :: v0_matrix, b0_next_matrix, b1_next_matrix, q_paid_matrix
DOUBLE PRECISION, DIMENSION(y_num, y_num) :: trans_matrix, cdf_matrix
DOUBLE PRECISION, DIMENSION(r_vol_num, r_vol_num) :: trans_matrix_r_vol, cdf_matrix_r_vol
DOUBLE PRECISION, DIMENSION(r_num, r_num, r_vol_num) :: trans_matrix_r, cdf_matrix_r
DOUBLE PRECISION, DIMENSION(y_num, r_num, r_vol_num) :: q_def_free_paid_matrix

contains

  subroutine discretize_AR(rho_ar, mu, sigma_eps, z, pi, w)
  implicit none

    !##### INPUT/OUTPUT VARIABLES #############################################
    ! autoregression parameter
    real*8, intent(in) :: rho_ar
    ! unconditional mean of the process
    real*8, intent(in) :: mu
    ! variance of the shock
    real*8, intent(in) :: sigma_eps
    ! discrete shock values
    real*8, intent(out) :: z(:)
    ! transition matrix
    real*8, intent(out) :: pi(:, :)
    ! the stationary distribution
    real*8, intent(out), optional :: w(:)
    !##### OTHER VARIABLES ####################################################
    integer :: n, in
    real*8 :: psi, sigma_eta
    !##### ROUTINE CODE #######################################################
    ! assert size equality and get approximation points
    ! n = assert_eq(size(z), size(pi,1), size(pi,2), 'discretize_AR') !-- This is the line in the CE-fortran toolbox
    n = size(z)
    ! calculate variance of the overall process
    sigma_eta = sigma_eps/(1d0-rho_ar**2d+0)
    ! determine the transition matrix
    call rouwenhorst_matrix(rho_ar, pi)
    ! determine the nodes
    psi = sqrt(dble(n-1))*sqrt(sigma_eta)
    do in = 1, n
        z(in) = -psi + 2d0*psi*dble(in-1)/dble(n-1)
    enddo
    z = z + mu
    if(present(w))then
        w = 1d0/dble(n)
        do in = 1, 10000
            w = matmul(transpose(pi), w)
        enddo
    endif
    !##########################################################################
    ! Subroutines and functions
    !##########################################################################
    contains
    !##########################################################################
    ! subroutine rouwenhorst_matrix
    !
    ! Calculates value of function that should be integrated for pis.
    !##########################################################################
    recursive subroutine rouwenhorst_matrix(rho_ar, pi_new)
        implicit none
        real*8, intent(in) :: rho_ar
        real*8, intent(out) :: pi_new(:, :)
        integer :: n
        real*8 :: p, pi_old(size(pi_new,1)-1, size(pi_new,1)-1)
        n = size(pi_new, 1)
        p = (1d0 + rho_ar)/2d0
        if(n == 2)then
            pi_new(1, :) = (/p, 1d0-p/)
            pi_new(2, :) = (/1d0-p, p/)
        else
            call rouwenhorst_matrix(rho_ar, pi_old)
            pi_new = 0d0
            pi_new(1:n-1, 1:n-1) = pi_new(1:n-1, 1:n-1) + p*pi_old
            pi_new(1:n-1, 2:n  ) = pi_new(1:n-1, 2:n  ) + (1d0-p)*pi_old
            pi_new(2:n  , 1:n-1) = pi_new(2:n  , 1:n-1) + (1d0-p)*pi_old
            pi_new(2:n  , 2:n  ) = pi_new(2:n  , 2:n  ) + p*pi_old
            pi_new(2:n-1, :) = pi_new(2:n-1, :)/2d0
        endif
    end subroutine

  end subroutine discretize_AR

end module param


!DEFINE UTILITY FUNCTION
DOUBLE PRECISION function u_fun(c)
USE param
DOUBLE PRECISION, INTENT(IN) :: c
DOUBLE PRECISION :: c_min, cons

c_min = 1d-10
cons = MAX(c,c_min)

if (sigma == 1) then
   u_fun = LOG(cons)
else
   u_fun = (cons**(1-sigma) ) / (1-sigma)
end if
end function u_fun


!SPECIFY GRID VALUES
subroutine compute_grid
USE param
implicit none
INTEGER :: i, j, k
DOUBLE PRECISION ::delta_y, prob_mass, y_pointer, qq, &
pdf_tirar, cdf_right, cdf_left,  prob_vector(y_num), y_left, y_right,&
prob_vector_r_vol(r_vol_num), prob_vector_r(r_num,r_vol_num), variance_r_vol, variance_r_innovation, scale_int, mu_sigma

!Discretize the s.s. for volatility shocks:

variance_r_vol = (std_u_vol*eta)**(2d+0)

if(r_vol_num < 2) then
   r_vol_grid = mean_r_vol
   trans_matrix_r_vol = 1d+0
   cdf_matrix_r_vol= 1d+0
   prob_vector_r_vol= 1d+0

else

   call discretize_AR(rho_r_vol, mean_r_vol , variance_r_vol, r_vol_grid, trans_matrix_r_vol, prob_vector_r_vol)

   do i=1,r_vol_num
      cdf_matrix_r_vol(i,1) = trans_matrix_r_vol(i,1)
      do j=2,r_vol_num
         cdf_matrix_r_vol(i,j) = cdf_matrix_r_vol(i,j-1) + trans_matrix_r_vol(i,j)
      end do
   end do
end if

open (1, FILE='graphs_r_vol_grid_dss.txt',STATUS='replace')
   WRITE(1, '(F12.8)') r_vol_grid
close(1)

open (100, FILE='graphs_trans_matrix_r_vol.txt',STATUS='replace')
do i=1,r_vol_num
   do j=1,r_vol_num
      WRITE(100,*) trans_matrix_r_vol(i,j)
   end do
end do
close(100)

open (2, FILE='graphs_pdf_r_vol.txt',STATUS='replace')
WRITE(2, '(F12.8)') prob_vector_r_vol
close(2)

!Next, do it for r_level:
if(r_vol_num < 2) then
   if(r_num < 2) then
      r_grid = zero
      trans_matrix_r = 1d+0
      cdf_matrix_r= 1d+0
      prob_vector_r= 1d+0

      variance_r_level_epsilon_tomorrow = 0d+0
   else
      
      scale_int = exp(0.5d+0 * eta**(2d+0)/(1d+0 - rho_r_vol**(2d+0)))
      variance_r_innovation = (std_u_r * exp(mean_r_vol) * scale_int)**(2d+0)
      ! scale_int is the fix for the Jensen's inequality (acknowledgement to a referee for poiting this out)
      call discretize_AR(rho_r, zero , variance_r_innovation, r_grid(:, 1), trans_matrix_r(:,:,1), prob_vector_r(:,1))

      variance_r_level_epsilon_tomorrow = variance_r_innovation
      !variance_r_level_epsilon_tomorrow is the conditional variance of innovations to the world interest rate *tomorrow*, conditional on
      !observing the \sigma_r today.

   end if

else

   do i=1,r_vol_num
      variance_r_innovation = (std_u_r * exp(r_vol_grid(i)))**(2)
      ! This is ok, because it's the conditional variance of the innovations to epsilon_r in period t, conditional on observing a realization 
      ! of the volatility in that period (r_vol_grid(i) .or. \sigma_{r,t} using the model's notation).
   
      call discretize_AR(rho_r, zero , variance_r_innovation, r_grid(:, i), trans_matrix_r(:,:,i), prob_vector_r(:,i))
   end do
   
   do i=1,r_vol_num
      
      mu_sigma = exp((1d+0-rho_r_vol)*mean_r_vol + rho_r_vol *r_vol_grid(i) + 0.5d+0 * eta**(2))
      variance_r_level_epsilon_tomorrow(i) = mu_sigma**(2)* exp(2d+0*eta)
   end do
   !variance_r_level_epsilon_tomorrow is the conditional variance of innovations to the world interest rate *tomorrow*, conditional on
   !observing the \sigma_r today.
end if

open (1, FILE='graphs_r_grid_dss.txt',STATUS='replace')
do k=1,r_vol_num
   do i=1,r_num
      WRITE(1, *) r_grid(i, k)
   end do
end do
close(1)

open (101, FILE='graphs_trans_matrix_r.txt',STATUS='replace')
do k=1,r_vol_num
   do i=1,r_num
      do j=1,r_num
         WRITE(101,*) trans_matrix_r(i,j,k)
      end do
   end do
end do
close(101)

do k=1,r_vol_num
   do i=1,r_num
      cdf_matrix_r(i,1,k) = trans_matrix_r(i,1,k)
      do j=2,r_num
         cdf_matrix_r(i,j,k) = cdf_matrix_r(i,j-1,k) + trans_matrix_r(i,j,k)
      end do
   end do
end do

open (3, FILE='graphs_pdf_r.txt',STATUS='replace')

do k=1,r_vol_num
      do j=1,r_num
         WRITE(3,*) prob_vector_r(j,k)
      end do
end do
close(3)


std_y = std_eps/ SQRT(1 - rho**2)
y_inf = mean_y - 4*std_y
y_sup = mean_y + 4*std_y
delta_y = 4*std_y / (y_num - 1d+0)

do i=1,y_num
   y_grid(i) = y_inf + (y_sup - y_inf) * (i-1) / (y_num - 1)
end do

open (1, FILE='graphs_y_grid_dss.txt',STATUS='replace')
 WRITE(1, '(F12.8)') y_grid
close(1)

do i=1,y_num

   y_pointer = (y_sup + delta_y - (1-rho)*mean_y - rho*y_grid(i))/ std_eps
   call nprob(y_pointer, cdf_right, qq, pdf_tirar)

   y_pointer = (y_inf - delta_y - (1-rho)*mean_y - rho*y_grid(i))/ std_eps
   call nprob(y_pointer, cdf_left, qq, pdf_tirar)

   prob_mass = cdf_right - cdf_left

   do j=1,y_num

      y_pointer = (y_grid(j) + delta_y - (1-rho)*mean_y - rho*y_grid(i))/ std_eps
      call nprob(y_pointer, cdf_right, qq, pdf_tirar)

      y_pointer = (y_grid(j) - delta_y - (1-rho)*mean_y - rho*y_grid(i))/ std_eps
      call nprob(y_pointer, cdf_left, qq, pdf_tirar)

      trans_matrix(i,j) = (cdf_right - cdf_left )/prob_mass

   end do

end do

open (100, FILE='graphs_trans_matrix_y.txt',STATUS='replace')
do i=1,y_num
   do j=1,y_num
      WRITE(100,*) trans_matrix(i,j)
   end do
end do
close(100)

do i=1,y_num
   cdf_matrix(i,1) = trans_matrix(i,1)
   do j=2,y_num
      cdf_matrix(i,j) = cdf_matrix(i,j-1) + trans_matrix(i,j)
   end do
end do

!1. Aproximate unconditional p.d.f. of y
open (1, FILE='graphs_pdf_y.txt',STATUS='replace')

y_left  = (y_grid(1) + delta_y)/ std_y
y_right = (y_grid(y_num) - delta_y)/ std_y

call nprob(y_left, cdf_left, qq, pdf_tirar)
call nprob(y_right, cdf_right, qq, pdf_tirar)

prob_vector(1)     = cdf_left
prob_vector(y_num) = 1d+0 - cdf_right

write(1, '(F12.8)') prob_vector(1)
   do j=2,y_num-1
     y_left  = (y_grid(j) - delta_y )/ std_y
     y_right = (y_grid(j) + delta_y )/ std_y

     call nprob(y_left, cdf_left, qq, pdf_tirar)
     call nprob(y_right, cdf_right, qq, pdf_tirar)


     prob_vector(j) = (cdf_right - cdf_left) !/prob_may_num

     write(1, '(F12.8)') prob_vector(j)

   end do
write(1, '(F12.8)') prob_vector(y_num)
close (1)

!Create grid for the assets (negative assets == debt)
b_inf = - 2.5d+0
b_sup =  0.0d+0

open (10, FILE='graphs_b_grid_dss.txt',STATUS='replace')

do i=1,b_num
   b_grid(i) = b_inf + (b_sup - b_inf) * (i-1) / (b_num - 1)
   WRITE(10, '(F12.8)') b_grid(i)
end do
close(10)

i_b_zero = b_num ! this is a fix for the grid specification that has no positive values.

!Create the default (indictor) grid
default_grid(1) = 0d+00
default_grid(2) = 1d+00

!Store the bounds and the value of delta (decay coupon)
open (13, FILE='graphs_bounds_dss.txt',STATUS='replace')
WRITE(13,*) b_inf, b_sup
WRITE(13,*) y_grid(1), y_grid(y_num)
WRITE(13,*) r_vol_grid(1), r_vol_grid(r_vol_num)
close(13)

open (14, FILE='graphs_delta.txt',STATUS='replace')
WRITE(14,*) delta
close(14)

end subroutine compute_grid

!ZERO-PROFIT BOND PRICE FUNCTION
DOUBLE PRECISION function q_def_free_fun(i_y, i_r, i_r_vol)
USE param
INTEGER, INTENT(IN) :: i_y, i_r, i_r_vol
DOUBLE PRECISION :: exp_m, exp_q, y_next, m_next, coupon_next, argument
INTEGER :: i, i_r_vol_next, i_r_next

exp_m = 0d+0
exp_q = 0d+0

do i_r_vol_next = 1,r_vol_num
   do i_r_next = 1,r_num
      do i=1,y_num
         
         y_next = y_grid(i)

         argument = r_prime + r_grid(i_r_next, i_r_vol_next) + &
         kappa * (y_next - ( (1d+0-rho)*mean_y + rho* y_initial)) + 0.5d+0*(kappa**2)*std_eps**2 + &
         kappa_rw * variance_r_level_epsilon_tomorrow(i_r_vol)

         m_next = EXP(- argument) 
         coupon_next = (r_prime + delta )/ (1d+0 + r_prime)
         
         exp_m   = exp_m + m_next * coupon_next * trans_matrix_r_vol(i_r_vol, i_r_vol_next)*&
        trans_matrix_r(i_r, i_r_next, i_r_vol_next)* trans_matrix(i_y, i)
        
        exp_q = exp_q + (1d+0 - delta) * m_next * q_def_free_paid_matrix(i, i_r_next, i_r_vol_next) * &
        trans_matrix_r_vol(i_r_vol, i_r_vol_next)* trans_matrix_r(i_r, i_r_next, i_r_vol_next)*&
        trans_matrix(i_y, i)
      end do
   end do
end do

q_def_free_fun = exp_m + exp_q

end function q_def_free_fun

!ZERO-PROFIT BOND PRICE FUNCTION
DOUBLE PRECISION function q_fun(i_b, i_y, i_r, i_r_vol)
USE param
INTEGER, INTENT(IN) :: i_b, i_y, i_r, i_r_vol
DOUBLE PRECISION :: exp_m, exp_q, y_next, m_next, coupon_next, argument
INTEGER :: i, i_r_vol_next, i_r_next

exp_m = 0d+0
exp_q = 0d+0

do i_r_vol_next = 1,r_vol_num
   do i_r_next = 1,r_num
      do i=1,y_num
         y_next = y_grid(i)

         argument = r_prime + r_grid(i_r_next, i_r_vol_next) + &
         kappa * ( y_next - ( (1d+0-rho)*mean_y + rho* y_initial)) + 0.5d+0*(kappa**2)*std_eps**2 + &
         kappa_rw * variance_r_level_epsilon_tomorrow(i_r_vol)

         m_next = EXP(- argument) 
         coupon_next = (r_prime + delta )/ (1d+0 + r_prime)

         exp_m   = exp_m + m_next * coupon_next * trans_matrix_r_vol(i_r_vol, i_r_vol_next)*&
         trans_matrix_r(i_r, i_r_next, i_r_vol_next)* trans_matrix(i_y, i)*&
         (1d+0 -default_grid(default_decision(i_b, i, i_r_next, i_r_vol_next )))

         exp_q   = exp_q + (1d+0 - delta) * m_next * q_paid_matrix(i_b, i, i_r_next, i_r_vol_next) * &
         trans_matrix_r_vol(i_r_vol, i_r_vol_next)* trans_matrix_r(i_r, i_r_next, i_r_vol_next)*&
         trans_matrix(i_y, i)* (1d+0 -default_grid(default_decision(i_b, i, i_r_next, i_r_vol_next )))
      end do
   end do
end do

q_fun = exp_m + exp_q
end function q_fun



!ZERO-PROFIT BOND PRICE FUNCTION
DOUBLE PRECISION function q_fun_old(i_b, i_y, i_r, i_r_vol)
USE param
INTEGER, INTENT(IN) :: i_b, i_y, i_r, i_r_vol
DOUBLE PRECISION :: partial_integral, prob_default
INTEGER :: i, i_r_vol_next, i_r_next

prob_default= 0d+0

do i_r_vol_next = 1,r_vol_num
   do i_r_next = 1,r_num
      do i=1,y_num
         prob_default  = prob_default  + trans_matrix_r_vol(i_r_vol, i_r_vol_next)*&
         trans_matrix_r(i_r, i_r_next, i_r_vol_next)* trans_matrix(i_y, i)*&
         default_grid(default_decision(i_b, i, i_r_next, i_r_vol_next ))

      end do
   end do
end do

! Tauchen's method may give room for tiny inaccuracies and the prob_default may exceed 1d+0,
!it may become a number like 1.000000004, so if that happens, we'll just assign it to 1d+0

prob_default = min(prob_default , 1d+0)

partial_integral =  1d+0 - prob_default

q_fun_old = partial_integral / (1d+0 + r_prime + r_grid(i_r, i_r_vol))

end function q_fun_old


!Compute the objective function in the Belman equation
DOUBLE PRECISION function objective_function(b)
USE param
DOUBLE PRECISION, INTENT(IN) :: b
INTEGER :: d
DOUBLE PRECISION ::  exp_v_next, exp_v_excl, q, output, consumption
DOUBLE PRECISION, EXTERNAL:: u_fun, q_fun

exp_v_next = exp_v_vector(i_b_next)      !0d+0
exp_v_excl = exp_v_excl_vector(i_b_next) !0d+0

d  = i_default_global
q = q_vector(i_b_next)

output = (1d+0 - default_grid(i_excl_global)) * EXP(y_initial) + &
default_grid(i_excl_global) * (EXP(y_initial)  - max(zero, d0*EXP(y_initial) + d1*EXP(y_initial)**2))


consumption = output - gov_spending +  &
(b_initial *coupon - q * (b-(1d+0 - delta)*b_initial) )* (1d+0-default_grid(i_excl_global))

objective_function = u_fun(consumption) + beta * mean_g**(1d+0 -sigma) *&
( default_grid(i_excl_global) * (prob_excl_end *exp_v_next + (1d+0 -prob_excl_end) * exp_v_excl) + &
(1d+0 - default_grid(i_excl_global)) * exp_v_next)

end function objective_function


! GET THE OPTIMAL B_NEXT, GIVEN INITIAL DEBT LEVEL AND INCOME LEVEL
subroutine optimize(index_opt, v_value)
USE param
integer ::i, index_opt, index_vector(1)
DOUBLE PRECISION :: b_next, v_value, vector(b_num)
DOUBLE PRECISION, EXTERNAL:: objective_function

do i=1,b_num
   i_b_next = i
   b_next   = b_grid(i)
   vector(i)  = objective_function(b_next)

end do

index_vector = maxloc(vector)
v_value = vector(index_vector(1))
index_opt = index_vector(1)

end subroutine optimize


subroutine iterate
USE param
DOUBLE PRECISION :: v_valor, convergence, criteria, deviation,deviation_q, b_next, &
v1_matrix_new(y_num,r_num,r_vol_num), exp_v_excl, exp_v_next, cons1_matrix(y_num,r_num,r_vol_num), &
q_def_free_paid_matrix_new(y_num,r_num,r_vol_num)
DOUBLE PRECISION, DIMENSION(b_num, y_num, r_num,r_vol_num) :: v0_matrix_new, cons0_matrix, cons_matrix, q_paid_matrix_new
DOUBLE PRECISION, DIMENSION(b_num, y_num, r_num,r_vol_num) :: v_matrix_new, dev_matrix, dev_matrix_q
INTEGER i_b, i_y, i_r, i_r_vol, i_r_next, i_r_vol_next, i_b_optimal0, i_b_optimal1, i_y_next, i_b_prime
DOUBLE PRECISION, EXTERNAL:: q_fun, objective_function, q_def_free_fun

criteria = 1.0d-3   !CRITERIA FOR CONVERGENCE -- Increase or decrease accordingly
convergence = -1

do WHILE(convergence<0)
   deviation = 0
   deviation_q = 0
   counter = counter + 1

   do i_r_vol = 1,r_vol_num
      do i_r = 1,r_num

         coupon = (r_prime + delta )/ (1d+0 + r_prime)

         do i_y = 1,y_num
            i_y_global = i_y
            y_initial = y_grid(i_y)
            
            ! HERE WE COMPUTE THE EXPECTED VALUE FUNCTIONS, AND ALSO THE PRICE FOR EACH POSSIBLE LEVEL OF B_NEXT
            do i_b_prime = 1,b_num
               exp_v_excl = 0d+0
               exp_v_next = 0d+0
               do i_r_vol_next =1,r_vol_num
                  do i_r_next = 1,r_num
                     do i_y_next=1,y_num
                        exp_v_next = exp_v_next + trans_matrix_r_vol(i_r_vol, i_r_vol_next)*&
                        trans_matrix_r(i_r, i_r_next, i_r_vol_next)*trans_matrix(i_y_global, i_y_next)*&
                        v_matrix(i_b_prime, i_y_next, i_r_next, i_r_vol_next)

                        exp_v_excl = exp_v_excl + trans_matrix_r_vol(i_r_vol, i_r_vol_next)*&
                        trans_matrix_r(i_r, i_r_next, i_r_vol_next)* trans_matrix(i_y_global, i_y_next)*&
                        v1_matrix(i_y_next, i_r_next, i_r_vol_next)
                     end do
                  end do
               end do

               exp_v_vector(i_b_prime) = exp_v_next
               exp_v_excl_vector(i_b_prime) = exp_v_excl
               q_vector(i_b_prime) = q_fun(i_b_prime, i_y_global, i_r, i_r_vol)

            end do

            ! NOW, COMPUTE THE VALUE FUNCTION OF DEFAULT (WHICH DEPENDS ONLY ON INCOME)

            i_default_global=2 !Country defaults
            i_excl_global=2

            i_b_next = i_b_zero
            i_b_optimal1 = i_b_zero
            b_next   = b_grid(i_b_next)
            v1_matrix_new(i_y, i_r, i_r_vol) = objective_function(b_next)
            cons1_matrix(i_y, i_r, i_r_vol) = EXP(y_initial) - gov_spending - &
            max(zero, d0*EXP(y_initial) + d1*EXP(y_initial)**2)
            q_def_free_paid_matrix_new(i_y, i_r, i_r_vol) = q_def_free_fun(i_y,i_r,i_r_vol)    

            do i_b = 1, b_num
               i_b_global = i_b
               b_initial = b_grid(i_b)

               ! NOW, COMPUTE THE VALUE FUNCTION OF REPAYMENT (WHICH DEPENDS ON DEBT AND INCOME)

               i_default_global= 1   ! Country does not default and is not excluded for the next period
               i_excl_global= 1
               ! THIS STEP FINDS THE OPTIMAL B_NEXT FOR A GIVEN INITIAL DEBT AND INCOME REALIZATION:
               call optimize(i_b_optimal0, v_valor)

               b0_next_matrix(i_b, i_y, i_r, i_r_vol) = b_grid(i_b_optimal0)
               b0_next_matrix_indice(i_b, i_y, i_r, i_r_vol) = i_b_optimal0

               v0_matrix_new(i_b, i_y, i_r, i_r_vol) = v_valor

               cons0_matrix(i_b,i_y, i_r, i_r_vol) = EXP(y_initial) - gov_spending + b_initial * coupon -&
               (b_grid(i_b_optimal0)-(1d+0 - delta)*b_initial)* q_fun(i_b_optimal0, i_y, i_r, i_r_vol)

               ! NOW, CHECK IF FOR THIS COMBINATION (b_initial,y_initial, i_r, i_r_vol) IT IS OPTIMAL TO DEFAULT OR NOT
               if (v1_matrix_new(i_y, i_r, i_r_vol) > v0_matrix_new( i_b, i_y, i_r, i_r_vol)) then
                  default_decision(i_b, i_y, i_r, i_r_vol) = 2
                  b_next_matrix(i_b, i_y, i_r, i_r_vol) = i_b_optimal1
                  v_matrix_new(i_b, i_y, i_r, i_r_vol) = v1_matrix_new(i_y, i_r, i_r_vol)
                  cons_matrix(i_b, i_y, i_r, i_r_vol) = cons1_matrix(i_y, i_r, i_r_vol)
                  q_paid_matrix_new(i_b, i_y, i_r, i_r_vol) =  0d+0 

               else
                  default_decision( i_b, i_y, i_r, i_r_vol) = 1
                  b_next_matrix(i_b, i_y, i_r, i_r_vol) = i_b_optimal0  
                  v_matrix_new(i_b, i_y, i_r, i_r_vol) = v0_matrix_new(i_b,i_y, i_r, i_r_vol)
                  cons_matrix(i_b, i_y, i_r, i_r_vol) = cons0_matrix(i_b,i_y, i_r, i_r_vol)
                  q_paid_matrix_new(i_b, i_y, i_r, i_r_vol) =  q_fun( i_b_optimal0, i_y, i_r, i_r_vol)

               end if
               !NOW, COMPUTE THE DEVIATION (I.E. HOW FAR WE ARE FROM CONVERGING):
               deviation = MAX(deviation, ABS(v_matrix_new( i_b, i_y, i_r, i_r_vol) - v_matrix(i_b, i_y, i_r, i_r_vol)))
               
               deviation_q = MAX(deviation_q, &
               ABS(q_paid_matrix_new( i_b, i_y, i_r, i_r_vol) - q_paid_matrix(i_b, i_y, i_r, i_r_vol)))

               dev_matrix(i_b, i_y, i_r, i_r_vol) = ABS(v_matrix_new(i_b, i_y, i_r, i_r_vol)-v_matrix( i_b, i_y, i_r, i_r_vol))
               
               dev_matrix_q(i_b, i_y, i_r, i_r_vol) = &
               ABS(q_paid_matrix_new(i_b, i_y, i_r, i_r_vol)-q_paid_matrix( i_b, i_y, i_r, i_r_vol))


            end do
         end do
      end do
   end do
   WRITE(*, *) counter, deviation, deviation_q

   !UPDATE VALUES OF MATRICES
   v0_matrix = v0_matrix_new
   v1_matrix = v1_matrix_new
   v_matrix = v_matrix_new
   q_paid_matrix = q_paid_matrix_new
   q_def_free_paid_matrix = q_def_free_paid_matrix_new

   ! FINALLY, CHECK WHETHER THE NEW DEVIATION IS SMALLER THAN THE TOLERANCE LEVEL (CALLED CRITERIA IN THIS CODE)
   if (deviation < criteria .or. counter > 1499) then 
    ! Here save things: 

      open (10, FILE='graphs_v_dss.txt',STATUS='replace')
      open (11, FILE='graphs_default_dss.txt',STATUS='replace')
      open (12, FILE='graphs_q_dss.txt',STATUS='replace')
      open (13, FILE='graphs_b_next_dss.txt',STATUS='replace')
      open (14, FILE='graphs_dev_dss.txt',STATUS='replace')
      open (16, FILE='graphs_q_paid_dss.txt',STATUS='replace')
      open (17, FILE='graphs_b0_next_dss.txt',STATUS='replace')
      open (18, FILE='graphs_consumption_dss.txt',STATUS='replace')
      open (19, FILE='graphs_counter.txt', STATUS='replace')
      open (20, FILE='graphs_q_def_free_dss.txt',STATUS='replace')

      WRITE(19, *) counter

      do i_r_vol = 1,r_vol_num
         do i_r = 1,r_num
            do i_y = 1,y_num
            
               i_y_global = i_y
               y_initial = y_grid(i_y)
               do i_b = 1,b_num
               
                  b_initial = b_grid(i_b)
                  i_b_global = i_b
               
                  WRITE(10,*) v_matrix_new(i_b, i_y,i_r, i_r_vol),&
                  v0_matrix_new(i_b, i_y, i_r, i_r_vol), v1_matrix_new( i_y, i_r, i_r_vol)
               
                  WRITE(11, '(F6.2)') default_grid(default_decision(i_b, i_y, i_r, i_r_vol))
                              
                  WRITE(12, '(F15.11)') q_fun(i_b, i_y, i_r, i_r_vol)
               
                  WRITE(13, *) b_next_matrix(i_b, i_y, i_r, i_r_vol)
                  
                  WRITE(17,*) b0_next_matrix(i_b, i_y, i_r, i_r_vol),&
                  b0_next_matrix_indice(i_b, i_y, i_r, i_r_vol)
            
                  WRITE(14, *) dev_matrix(i_b, i_y, i_r, i_r_vol), dev_matrix_q(i_b, i_y, i_r, i_r_vol)
                  
                  WRITE(16, '(F15.11)') q_paid_matrix_new(i_b, i_y, i_r, i_r_vol)
                  
                  WRITE(18,*) cons_matrix(i_b, i_y,i_r, i_r_vol), cons0_matrix(i_b, i_y, i_r, i_r_vol),&
                  cons1_matrix( i_y, i_r, i_r_vol)
                  
                  WRITE(20, *) q_def_free_paid_matrix_new(i_y, i_r, i_r_vol)

               end do
            end do
         end do
      end do
      CLOSE(10)
      CLOSE(11)
      CLOSE(12)
      CLOSE(13)
      CLOSE(14)
      CLOSE(16)
      CLOSE(17)
      CLOSE(18)
      CLOSE(19)      
      CLOSE(20)      

      ! Declare success:
      convergence =1
   end if
end do

end subroutine iterate

subroutine solve_q_nodef
USE param
DOUBLE PRECISION :: dev, criterion_q_nodef, q_nodef_new(y_num,r_num,r_vol_num), coupon_nodef,&
SDF_matrix(y_num,r_num, r_vol_num, y_num, r_num, r_vol_num), argument
INTEGER :: i_y, i_r, i_r_vol, i_y_next, i_r_vol_next, i_r_next

! 1. First, compute the SDF_matrix
do i_r_vol = 1,r_vol_num
   do i_r = 1,r_num
      do i_y = 1,y_num
         y_initial = y_grid(i_y)

         do i_r_vol_next = 1,r_vol_num
            do i_r_next = 1,r_num
               do i_y_next = 1,y_num

                  argument = (r_prime + r_grid(i_r_next, i_r_vol_next)) + &
                  kappa * ( y_grid(i_y_next) - ( (1d+0-rho)*mean_y + rho* y_initial)) + 0.5d+0*(kappa**2)*std_eps**2 + &
                  kappa_rw * variance_r_level_epsilon_tomorrow(i_r_vol)

                  SDF_matrix(i_y_next,i_r_next,i_r_vol_next,i_y,i_r,i_r_vol)= EXP(-argument) 

               end do
            end do
         end do
      end do
   end do
end do

!1.5 Intermediate step: compute the Expected SDF (I don't need it here, but I wanted for the simulations)
expected_SDF_matrix(i_y, i_r, i_r_vol) = 0d+0

do i_r_vol = 1,r_vol_num
   do i_r = 1,r_num
      do i_y = 1,y_num
         y_initial = y_grid(i_y)
         do i_r_vol_next = 1,r_vol_num
            do i_r_next = 1,r_num
               do i_y_next = 1,y_num
                 expected_SDF_matrix(i_y, i_r, i_r_vol) = expected_SDF_matrix(i_y, i_r, i_r_vol) + &
                 SDF_matrix(i_y_next,i_r_next,i_r_vol_next,i_y,i_r, i_r_vol) *&
                 trans_matrix_r_vol(i_r_vol, i_r_vol_next)* trans_matrix_r(i_r, i_r_next, i_r_vol_next)*&
                 trans_matrix(i_y, i_y_next)
               end do
            end do
         end do
      end do
   end do
end do


!2. Iterate on the q_nodef

dev = 10d+0
criterion_q_nodef = 1d-6
q_nodef = 0d+0

do while(dev > criterion_q_nodef)
   q_nodef_new = 0d+0
   dev = 0d+0
   !calculate long-term bond price

   do i_r_vol = 1,r_vol_num
      do i_r = 1,r_num
         do i_y = 1,y_num
            y_initial = y_grid(i_y)
            do i_r_vol_next = 1,r_vol_num
               do i_r_next = 1,r_num
                  do i_y_next = 1,y_num                     
                     coupon_nodef = (r_prime + delta )/ (1d+0 + r_prime)
                     
                     q_nodef_new(i_y, i_r, i_r_vol) = q_nodef_new(i_y, i_r, i_r_vol) + &
                     SDF_matrix(i_y_next,i_r_next,i_r_vol_next,i_y,i_r, i_r_vol) *&
                     (coupon_nodef+(1d+0-delta)*q_nodef(i_y_next,i_r_next,i_r_vol_next))*&
                     trans_matrix_r_vol(i_r_vol, i_r_vol_next)*&
                     trans_matrix_r(i_r, i_r_next, i_r_vol_next)*trans_matrix(i_y, i_y_next)
                  end do
               end do
            end do
            dev = max(abs(q_nodef_new(i_y,i_r,i_r_vol)-q_nodef(i_y,i_r,i_r_vol)), dev)

         end do
      end do
   end do
   q_nodef = q_nodef_new

end do

!3. Save q_nodef and SDF_matrix

open (11, FILE='graphs_sdf_matrix.txt',STATUS='replace')
open (12, FILE='graphs_q_def_free.txt',STATUS='replace')
open (13, FILE='graphs_expected_sdf_matrix.txt',STATUS='replace')

do i_r_vol = 1,r_vol_num
   do i_r = 1,r_num
      do i_y = 1,y_num
         write(12, *) q_nodef(i_y, i_r, i_r_vol)
         write(13, *) expected_SDF_matrix(i_y, i_r, i_r_vol)
      end do
   end do
end do

do i_r_vol = 1,r_vol_num
   do i_r = 1,r_num
      do i_y = 1,y_num
         do i_r_vol_next = 1,r_vol_num
            do i_r_next = 1,r_num
               do i_y_next = 1,y_num

                  write(11, *) SDF_matrix(i_y_next,i_r_next,i_r_vol_next,i_y,i_r, i_r_vol)

               end do
            end do
         end do
      end do
   end do
end do

close(11)
close(12)
close(13)

end subroutine solve_q_nodef

program main
USE param
DOUBLE PRECISION :: start_time, end_time, indicator_external
INTEGER  i_b, i_y, i_r, i_r_vol
character (20) :: name_1, name_2, name_3, name_4, name_5
DOUBLE PRECISION, EXTERNAL:: q_fun, u_fun

open (1, FILE='param_calib_beta.txt')
READ(1, *) name_1, beta
close (1)

open (2, FILE='param_calib_d0.txt')
READ(2, *) name_2, d0
close (2)
d0 = d0/100d+0

open (3, FILE='param_calib_d1.txt')
READ(3, *) name_3, d1
close (3)
d1 = d1/100d+0
 
open (4, FILE='param_calib_kappa.txt')
READ(4, *) name_4, kappa
close (4)
kappa = kappa/10d+0

open (5, FILE='param_calib_kappa_rw.txt')
READ(5, *) name_5, kappa_rw
close (5)
d0 = d0 - d1

if(d0 > 99d+0) then !d0_hat=100 is an ad-hoc value to induce what follows/
d0 =zero ! this is to have just a quad component in the cost
end if

write(nout, '(A12, X, F12.8)') 'beta  =', beta
write(nout, '(A12, X, F12.8)') 'd0 =', d0
write(nout, '(A12, X, F12.8)') 'd1 =', d1
write(nout, '(A12, X, F12.8)') 'kappa =', kappa
write(nout, '(A12, X, F12.8)') 'kappa_rw =', kappa_rw

call cpu_time(start_time)
call compute_grid
call solve_q_nodef

indicator_external = 0 !IF SET TO 1 OR LARGER, CODE WILL START FROM PREVIOUSLY STORED ITERATION

if (indicator_external < 0.5d+0) then
   !THIS PART IS THE COMPUTATION OF THE VALUE FUNCTIONS FOR THE LAST PERIOD
   ! --> THESE WILL BE THE INITIAL GUESSES FOR THE VFI ALGORITHM

   counter = 0

   do i_r_vol = 1,r_vol_num
      do i_r = 1,r_num
         coupon = (r_prime + delta) / (1d+0 + r_prime)
         
         do i_y = 1,y_num

            y_initial = y_grid(i_y)

            v1_matrix( i_y, i_r,i_r_vol) = u_fun( EXP(y_initial) - gov_spending  - &
            max(zero, d0*EXP(y_initial) + d1*EXP(y_initial)**2) )
            
            q_def_free_paid_matrix(i_y, i_r, i_r_vol) = 0d+0

            do i_b = 1,b_num

               b_initial = b_grid(i_b)

               v0_matrix(i_b, i_y, i_r,i_r_vol) = u_fun(EXP(y_initial) - gov_spending + b_grid(i_b) * coupon)

               if(v1_matrix(i_y, i_r,i_r_vol) > v0_matrix(i_b, i_y, i_r,i_r_vol) )then
                  default_decision(i_b,i_y, i_r,i_r_vol)=2
                  v_matrix(i_b, i_y, i_r,i_r_vol) = v1_matrix(i_y, i_r,i_r_vol)
               else
                  default_decision(i_b,i_y, i_r,i_r_vol)=1
                  v_matrix(i_b, i_y, i_r,i_r_vol) = v0_matrix(i_b,i_y, i_r,i_r_vol)
               end if

               q_paid_matrix(i_b, i_y, i_r, i_r_vol) = 0d+0
               
            end do
         end do
      end do
   end do

else !READ DATA FROM EXTERNAL FILES

   open (10, FILE='graphs_v_dss.txt')
   open (12, FILE='graphs_q_def_free.txt')
   open (13, FILE='graphs_b_next_dss.txt')
   open (16, FILE='graphs_q_paid_dss.txt')
   open (19, FILE='graphs_counter.txt')
   open (20, FILE='graphs_q_def_free_dss.txt')
   
   READ(19,*) counter
   write(*,*) counter

   do i_r_vol = 1,r_vol_num
      do i_r = 1,r_num
         do i_y = 1,y_num
            READ(12, *) q_nodef(i_y, i_r, i_r_vol)            
            READ(20, *) q_def_free_paid_matrix(i_y, i_r, i_r_vol)

            do i_b = 1,b_num
               READ(16, *) q_paid_matrix(i_b, i_y, i_r, i_r_vol)
               READ(13, *) b_next_matrix(i_b, i_y, i_r, i_r_vol)

               READ(10, *) v_matrix(i_b, i_y, i_r, i_r_vol), v0_matrix(i_b, i_y, i_r, i_r_vol), v1_matrix(i_y, i_r, i_r_vol)

               if (v0_matrix(i_b, i_y,i_r, i_r_vol)< v1_matrix(i_y, i_r, i_r_vol)) then
                  default_decision(i_b, i_y, i_r, i_r_vol) =2
               else
                  default_decision(i_b, i_y, i_r, i_r_vol) =1
               end if

            end do
         end do
      end do
   end do
      
   CLOSE(10)
   CLOSE(12)
   CLOSE(13)
   close (16)
   CLOSE(19)
   CLOSE(20)

end if

call iterate
call simulate
call irf

call cpu_time(end_time)
WRITE(*, *) 'Hours ', 'Minutes', 'Seconds'
WRITE(*, *) INT((end_time - start_time) / 3600d+0), &
INT((end_time-start_time)/60d+0 - INT((end_time - start_time) / 3600d+0)*60d+0),&
INT(end_time-start_time - INT((end_time - start_time) / 3600d+0)*3600d+0 - &
INT((end_time-start_time)/60d+0 - INT((end_time - start_time) / 3600d+0)*60d+0)*60d+0)

end program main


subroutine simulate
USE param
integer :: period_num, i,j, sample_num, default_num, i_y_previous, i_y_current, i_b_current,i_r_previous,&
     i_r_current,i_r_vol_previous,i_r_vol_current
parameter (sample_num =500, period_num=1501)
DOUBLE PRECISION :: random_matrix(period_num, sample_num, 2), random_vector(1:2*sample_num*period_num), &
      random_matrix_rw(period_num, 2), random_vector_rw(1:2*period_num), &
     z_shock(period_num,sample_num), r(period_num,sample_num), r_vol_matrix(period_num,sample_num), b(period_num+1,sample_num),&
     q(period_num,sample_num), qdef, b_zero, c(period_num,sample_num), tb(period_num,sample_num),&
     y(period_num,sample_num), q_default_free(period_num,sample_num), aux_real, q_default_free_new(period_num,sample_num)
INTEGER :: d(period_num, sample_num), excl(period_num,sample_num)
DOUBLE PRECISION, EXTERNAL:: q_fun, q_def_free_fun

b_zero = 0d+0  !IF THE COUNTRY IS NOT EXCLUDED AFTER A DEFAULT EPISODE ==> BORROWS AS IF IT STARTS WITH ZERO DEBT

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!! NOTE !!!!!!
! z DENOTE UNDERLYING SHOCK TO THE ENDOWMENT
! y DENOTE REALIZED ENDOWMENT (EXP(z))
! CODE WAS WRITEN WITH y = SHOCK, SO mean_y ACTUALLY = E(z)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

! NOTE: UNCOMMENT FOLLOWING BOLOCK IF YOU WANT TO USE IMSL'S ROUNTINES: RNSET AND DRNUN (TO GENERATE AN INITIAL SEED AND
! CALL A RANDOM NUMBER GENERATOR).
!ivalue=1234              !SEED

CALL init_random_seed()         ! see example of RANDOM_SEED
CALL RANDOM_NUMBER(random_vector)
CALL RANDOM_NUMBER(random_vector_rw) !important to always call these two lines in the same order.

!First column of random_matrix is used to generate y shocks.
!Second column of random_matrix is used to generate the reentry shocks.
do j=1,sample_num
   do i=1,period_num
      random_matrix(i,j,1)=random_vector((j-1)*period_num+i)
      random_matrix(i,j,2)=random_vector(period_num*sample_num + (j-1)*period_num+i)
   end do
end do

!First column of random_matrix is used to generate rw level shocks
!Second column of random_matrix is used to generate rw volatility shocks.
do i=1,period_num
   random_matrix_rw(i,1)=random_vector_rw(i)
   random_matrix_rw(i,2)=random_vector_rw(period_num+i)
end do

open (UNIT=21, FILE="graphs_data_sim_dss.txt", status = 'replace')
open (UNIT=22, FILE="graphs_def_per_dss.txt", status = 'replace')
open (UNIT=23, FILE="graphs_param_dss.txt", status = 'replace')
open (UNIT=24, FILE="graphs_tirar_dss.txt", status = 'replace')
open (UNIT=25, FILE="graphs_def_num_dss.txt", status = 'replace')

WRITE(23, '(I10)') period_num-1
WRITE(23, '(I10)') sample_num
CLOSE(23)

i_default_global = 1

do j=1,sample_num   !SOLVE FOR SAMPLE j
   WRITE(*, *) j
   aux_real = y_num/2d+0
   i_y_previous = floor(aux_real)
   
   if (r_vol_num < 2) then
      i_r_vol_previous = 1
      
      if(r_num < 2) then
         i_r_previous = 1
      else
         aux_real = r_num/2d+0
         i_r_previous = floor(aux_real)
      end if

   else

      aux_real = r_vol_num/2d+0
      i_r_vol_previous =floor(aux_real)

      aux_real = r_num/2d+0
      i_r_previous = floor(aux_real)
   end if


   z_shock(1,j) = y_grid(i_y_previous)
   r(1,j) = r_prime + r_grid(i_r_previous, i_r_vol_previous)
   r_vol_matrix(1,j) = r_vol_grid(i_r_vol_previous)
   y(1,j) = EXP(z_shock(1,j))
   b(1,j) = zero
   i_b_next = b_next_matrix(i_b_zero, i_y_previous, i_r_previous, i_r_vol_previous)
   b(2,j) = b_grid(i_b_next)
   d(1,j) = 1    !NO DEFAULT IN FIRST PERIOD
   excl(1,j) = 1 !NO EXCLUSION IN FIRST PERIOD
   default_num =0  !variable that counts the number of defaults in each sample.

   do i=2,period_num

      qdef = 0d+0

      !THE FOLLOWING 4 LINES ARE A WAY OF GETTING A RANDOM DRAW FOR THE INCOME SHOCK
      i_y_current = 1
      do WHILE(cdf_matrix(i_y_previous, i_y_current) < random_matrix(i,j,1))
         i_y_current = i_y_current + 1
      end do

      i_y_previous = i_y_current
      z_shock(i,j) = y_grid(i_y_current)

      !THE FOLLOWING 4 LINES ARE A WAY OF GETTING A RANDOM DRAW FOR THE INTEREST SHOCK
      i_r_vol_current = 1      
      do WHILE(cdf_matrix_r_vol(i_r_vol_previous, i_r_vol_current) < random_matrix_rw(i,2))
         i_r_vol_current = i_r_vol_current + 1
      end do
      i_r_vol_previous = i_r_vol_current

      i_r_current = 1
      do WHILE(cdf_matrix_r(i_r_previous, i_r_current,i_r_vol_current) < random_matrix_rw(i,1))
         i_r_current = i_r_current + 1
      end do
      i_r_previous = i_r_current
      r(i,j) = r_prime + r_grid(i_r_current, i_r_vol_current)
      r_vol_matrix(i,j) = r_vol_grid(i_r_vol_current)

      i_b_current = i_b_next  !CURRENT INDEX OF ASSET = PREVIOUS INDEX OF SAVING
      y_initial = z_shock(i,j)
      b_initial = b(i,j)

      if (excl(i-1,j) ==2 ) THEN  !COUNTRY WAS EXCLUDED IN THE PREVIOUS PERIOD

         if (random_matrix(i,j,2) < prob_excl_end) THEN  ! EXCLUSION ENDS THIS PERIOD
            d(i,j) = 1   !COUNTRY DOES NOT DEFAULT
            excl(i,j) = 1
            i_b_next = b_next_matrix(i_b_current, i_y_current, i_r_current, i_r_vol_current)
            b(i+1,j) = b_grid(i_b_next)
            q(i,j) = q_fun(i_b_next, i_y_current, i_r_current, i_r_vol_current)
            y(i,j) = EXP(z_shock(i,j))
            c(i,j) = y(i,j) - gov_spending  - q(i,j)*b(i+1,j)
            tb(i,j) = y(i,j) - c(i,j) - gov_spending
            q_default_free(i,j) = q_nodef(i_y_current, i_r_current, i_r_vol_current)
            q_default_free_new(i,j) = q_def_free_fun(i_y_current, i_r_current, i_r_vol_current)

         else ! EXCLUSION DOES NOT END THIS PERIOD
            d(i,j) = 1
            excl(i,j) = 2
            i_b_next = i_b_zero
            b(i+1,j) = b_grid(i_b_next)
            q(i,j) = q_fun(i_b_next, i_y_current, i_r_current, i_r_vol_current)
            y(i,j) = EXP(z_shock(i,j))  - max(zero, d0*EXP(z_shock(i,j)) + d1*EXP(z_shock(i,j))**2d+0)
            c(i,j) = y(i,j) - gov_spending
            tb(i,j) = y(i,j) - c(i,j)  - gov_spending
            q_default_free(i,j) = q_nodef(i_y_current, i_r_current, i_r_vol_current)
            q_default_free_new(i,j) = q_def_free_fun(i_y_current, i_r_current, i_r_vol_current)

         end if

      else !COUNTRY WAS NOT EXCLUDED IN THE PREVIOUS PERIOD

         if (default_decision(i_b_current, i_y_current, i_r_current, i_r_vol_current) > 1) then  !COUNTRY DEFAULTS
            d(i,j) = 2
            excl(i,j) = 2
            i_b_next = i_b_zero
            b(i+1,j) = b_grid(i_b_next)
            q(i,j) = q_fun(i_b_next, i_y_current, i_r_current, i_r_vol_current)  !q_paid
            y(i,j) = EXP(z_shock(i,j))  - max(zero, d0*EXP(z_shock(i,j)) + d1*EXP(z_shock(i,j))**2d+0)
            c(i,j) = y(i,j) - gov_spending
            tb(i,j) = y(i,j) - c(i,j) - gov_spending
            q_default_free(i,j) = q_nodef(i_y_current, i_r_current, i_r_vol_current)
            q_default_free_new(i,j) = q_def_free_fun(i_y_current, i_r_current, i_r_vol_current)
            qdef = zero
            WRITE(22, '(I7)') i-1 !NEED TO SUBSTRACT 1. REASON: files start saving data on period 2
            default_num = default_num + 1
          else
            d(i,j) = 1   !COUNTRY DOES NOT DEFAULT
            excl(i,j) = 1
            i_b_next = b_next_matrix(i_b_current, i_y_current, i_r_current, i_r_vol_current)
            b(i+1,j) = b_grid(i_b_next)

            q(i,j) = q_fun(i_b_next, i_y_current, i_r_current, i_r_vol_current)  !q_paid
            y(i,j) = EXP(z_shock(i,j))

            coupon = (r_prime+delta)/(1d+0 +r_prime)
            c(i,j) = y(i,j)  - gov_spending + b(i,j) * coupon - q(i,j)* (b(i+1,j) - b(i,j)*(1d+0 - delta))
            tb(i,j) = y(i,j) - c(i,j) - gov_spending
            q_default_free(i,j) = q_nodef(i_y_current, i_r_current, i_r_vol_current)
            q_default_free_new(i,j) = q_def_free_fun(i_y_current, i_r_current, i_r_vol_current)

         end if
      end if

      WRITE(21, *) y(i,j), b(i,j), q(i,j), qdef, c(i,j), tb(i,j)/y(i,j), excl(i,j), &
      d(i,j), r(i,j), q_default_free(i,j), i_y_current, i_r_current, i_r_vol_current, r_vol_matrix(i,j), &
      expected_SDF_matrix(i_y_current, i_r_current, i_r_vol_current), q_default_free_new(i,j)




   end do
   WRITE(25, *) default_num
end do

CLOSE(21)
CLOSE(22)
CLOSE(24)
CLOSE(25)

end subroutine simulate


!============================= UTILITIES =============================

SUBROUTINE init_random_seed()
  INTEGER ::  n
  INTEGER, DIMENSION(:), ALLOCATABLE :: seed

  CALL RANDOM_SEED(size = n)
  ALLOCATE(seed(n))

  seed = 12345
  !          write(*,*) seed,n

  CALL RANDOM_SEED(PUT = seed)

  DEALLOCATE(seed)
END SUBROUTINE init_random_seed


subroutine nprob (z, p, q, pdf)
  !*****************************************************************************80
  !
  !! NPROB computes the cumulative density of the standard normal distribution.
  !
  !  Modified:
  !
  !    13 January 2008
  !
  !  Author:
  !
  !    Original FORTRAN77 version by AG Adams.
  !    FORTRAN90 version by John Burkardt.
  !
  !  Reference:
  !
  !    AG Adams,
  !    Algorithm 39:
  !    Areas Under the Normal Curve,
  !    Computer Journal,
  !    Volume 12, Number 2, May 1969, pages 197-198.
  !
  !  Parameters:
  !
  !    Input, DOUBLE PRECISION Z, divides the real line into
  !    two semi-infinite intervals, over each of which the standard normal
  !    distribution is to be integrated.
  !
  !    Output, DOUBLE PRECISION P, Q, the integrals of the standard normal
  !    distribution over the intervals ( - Infinity, Z] and
  !    [Z, + Infinity ), respectively.
  !
  !    Output, DOUBLE PRECISION PDF, the value of the standard normal
  !    distribution at Z.
  !
  implicit none
  DOUBLE PRECISION, parameter :: a0 = 0.5D+00
  DOUBLE PRECISION, parameter :: a1 = 0.398942280444D+00
  DOUBLE PRECISION, parameter :: a2 = 0.399903438504D+00
  DOUBLE PRECISION, parameter :: a3 = 5.75885480458D+00
  DOUBLE PRECISION, parameter :: a4 = 29.8213557808D+00
  DOUBLE PRECISION, parameter :: a5 = 2.62433121679D+00
  DOUBLE PRECISION, parameter :: a6 = 48.6959930692D+00
  DOUBLE PRECISION, parameter :: a7 = 5.92885724438D+00
  DOUBLE PRECISION, parameter :: b0 = 0.398942280385D+00
  DOUBLE PRECISION, parameter :: b1 = 0.000000038052D+00
  DOUBLE PRECISION, parameter :: b2 = 1.00000615302D+00
  DOUBLE PRECISION, parameter :: b3 = 0.000398064794D+00
  DOUBLE PRECISION, parameter :: b4 = 1.98615381364D+00
  DOUBLE PRECISION, parameter :: b5 = 0.151679116635D+00
  DOUBLE PRECISION, parameter :: b6 = 5.29330324926D+00
  DOUBLE PRECISION, parameter :: b7 = 4.8385912808D+00
  DOUBLE PRECISION, parameter :: b8 = 15.1508972451D+00
  DOUBLE PRECISION, parameter :: b9 = 0.742380924027D+00
  DOUBLE PRECISION, parameter :: b10 = 30.789933034D+00
  DOUBLE PRECISION, parameter :: b11 = 3.99019417011D+00
  DOUBLE PRECISION :: p, pdf, q, y, z, zabs

  zabs = abs (z)
  !
  !  |Z| between 0 and 1.28
  !
  if ( abs ( z ) <= 1.28D+00 ) then

     y = a0 * z * z
     pdf = exp ( - y ) * b0

     q = a0 - zabs * ( a1 - a2 * y/( y + a3 - a4/( y + a5 + a6/( y + a7 ))))
     !
     !  |Z| between 1.28 and 12.7
     !
  else if ( abs ( z ) <= 12.7D+00 ) then

     y = a0 * z * z
     pdf = exp ( - y ) * b0

     q = pdf &
          / ( zabs - b1 + b2 &
          / ( zabs + b3 + b4 &
          / ( zabs - b5 + b6 &
          / ( zabs + b7 - b8 &
          / ( zabs + b9 + b10 &
          / ( zabs + b11 ))))))
     !
     !  Z far out in tail.
     !
  else

     q = 0.0D+00
     pdf = 0.0D+00

  end if

  if ( z < 0.0D+00 ) then
     p = q
     q = 1.0D+00 - p
  else
     p = 1.0D+00 - q
  end if

  return
end subroutine nprob



subroutine irf
USE param
integer :: period_num, i,j, sample_num, default_num, i_y_previous, i_y_current, i_b_current,i_r_previous,&
     i_r_current,i_r_vol_previous,i_r_vol_current
parameter (sample_num =500, period_num=1001)
DOUBLE PRECISION :: random_matrix(period_num, sample_num, 2), random_vector(1:2*sample_num*period_num), &
      random_matrix_rw(period_num, 2), random_vector_rw(1:2*period_num), &
      z_shock(period_num,sample_num), r(period_num,sample_num), r_vol_matrix(period_num,sample_num), b(period_num+1,sample_num),&
      q(period_num,sample_num), qdef, b_zero, c(period_num,sample_num), tb(period_num,sample_num),&
      y(period_num,sample_num), q_default_free(period_num,sample_num), aux_real, q_default_free_new(period_num,sample_num)
INTEGER :: d(period_num, sample_num), excl(period_num,sample_num)
DOUBLE PRECISION, EXTERNAL:: q_fun, q_def_free_fun

b_zero = 0d+0  !IF THE COUNTRY IS NOT EXCLUDED AFTER A DEFAULT EPISODE ==> BORROWS AS IF IT STARTS WITH ZERO DEBT

CALL init_random_seed()         ! see example of RANDOM_SEED
CALL RANDOM_NUMBER(random_vector)
CALL RANDOM_NUMBER(random_vector_rw) !important to always call these two lines in the same order.

!First column of random_matrix is used to generate y shocks.
!Second column of random_matrix is used to generate the reentry shocks.
do j=1,sample_num
   do i=1,period_num
      random_matrix(i,j,1)=random_vector((j-1)*period_num+i)
      random_matrix(i,j,2)=random_vector(period_num*sample_num + (j-1)*period_num+i)
   end do
end do

!First column of random_matrix is used to generate rw level shocks
!Second column of random_matrix is used to generate rw volatility shocks.
do i=1,period_num
   random_matrix_rw(i,1)=random_vector_rw(i)
   random_matrix_rw(i,2)=random_vector_rw(period_num+i)
end do


open (UNIT=21, FILE="graphs_data_irf.txt", status = 'replace')
open (UNIT=23, FILE="graphs_params_irf.txt", status = 'replace')

WRITE(23, '(I10)') period_num-1
WRITE(23, '(I10)') sample_num
CLOSE(23)

i_default_global = 1  !I DON'T THINK WE ARE USING THIS.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
do j=1,sample_num   !SOLVE FOR SAMPLE j
   WRITE(*, *) j
   aux_real = y_num/2d+0
   i_y_previous = floor(aux_real)
   
   if (r_vol_num < 2) then
      i_r_vol_previous = 1
      
      if(r_num < 2) then
         i_r_previous = 1
      else
         aux_real = r_num/2d+0
         i_r_previous = floor(aux_real)
      end if

   else

      aux_real = r_vol_num/2d+0
      i_r_vol_previous =floor(aux_real)

      aux_real = r_num/2d+0
      i_r_previous = floor(aux_real)
   end if


   z_shock(1,j) = y_grid(i_y_previous)
   r(1,j) = r_prime + r_grid(i_r_previous, i_r_vol_previous)
   r_vol_matrix(1,j) = r_vol_grid(i_r_vol_previous)
   y(1,j) = EXP(z_shock(1,j))
   b(1,j) = zero
   i_b_next = b_next_matrix(i_b_zero, i_y_previous, i_r_previous, i_r_vol_previous)
!   i_b_next = i_b_zero                   !INDEX OF BOND POSITION IN THE NEXT PERIOD.
   b(2,j) = b_grid(i_b_next)
   d(1,j) = 1    !NO DEFAULT IN FIRST PERIOD
   excl(1,j) = 1 !NO EXCLUSION IN FIRST PERIOD
   default_num =0  !variable that counts the number of defaults in each sample.

   do i=2,period_num

      qdef = 0d+0

      !THE FOLLOWING 4 LINES ARE A WAY OF GETTING A RANDOM DRAW FOR THE INCOME SHOCK
      i_y_current = 1
      do WHILE(cdf_matrix(i_y_previous, i_y_current) < random_matrix(i,j,1))
         i_y_current = i_y_current + 1
      end do

      i_y_previous = i_y_current
      z_shock(i,j) = y_grid(i_y_current)


      !THE FOLLOWING 4 LINES ARE A WAY OF GETTING A RANDOM DRAW FOR THE INTEREST SHOCK
      i_r_vol_current = 1
      do WHILE(cdf_matrix_r_vol(i_r_vol_previous, i_r_vol_current) < random_matrix_rw(i,2))
         i_r_vol_current = i_r_vol_current + 1
      end do
      i_r_vol_previous = i_r_vol_current

      i_r_current = 1
      do WHILE(cdf_matrix_r(i_r_previous, i_r_current,i_r_vol_current) < random_matrix_rw(i,1))
         i_r_current = i_r_current + 1
      end do
      i_r_previous = i_r_current
      r(i,j) = r_prime + r_grid(i_r_current, i_r_vol_current)
      r_vol_matrix(i,j) = r_vol_grid(i_r_vol_current)


      i_b_current = i_b_next  !CURRENT INDEX OF ASSET = PREVIOUS INDEX OF SAVING
      y_initial = z_shock(i,j)
      b_initial = b(i,j)


      if (excl(i-1,j) ==2 ) THEN  !COUNTRY WAS EXCLUDED IN THE PREVIOUS PERIOD

         if (random_matrix(i,j,2) < prob_excl_end) THEN  ! EXCLUSION ENDS THIS PERIOD
            d(i,j) = 1   !COUNTRY DOES NOT DEFAULT
            excl(i,j) = 1
            i_b_next = b_next_matrix(i_b_current, i_y_current, i_r_current, i_r_vol_current)
            b(i+1,j) = b_grid(i_b_next)
            q(i,j) = q_fun(i_b_next, i_y_current, i_r_current, i_r_vol_current)
            y(i,j) = EXP(z_shock(i,j))
            c(i,j) = y(i,j) - gov_spending  - q(i,j)*b(i+1,j)
            tb(i,j) = y(i,j) - c(i,j) - gov_spending
            q_default_free(i,j) = q_nodef(i_y_current, i_r_current, i_r_vol_current)
            q_default_free_new(i,j) = q_def_free_fun(i_y_current, i_r_current, i_r_vol_current)

         else ! EXCLUSION DOES NOT END THIS PERIOD
            d(i,j) = 1
            excl(i,j) = 2
            i_b_next = i_b_zero
            b(i+1,j) = b_grid(i_b_next)
            q(i,j) = q_fun(i_b_next, i_y_current, i_r_current, i_r_vol_current)
            y(i,j) = EXP(z_shock(i,j))  - max(zero, d0*EXP(z_shock(i,j)) + d1*EXP(z_shock(i,j))**2d+0)
           ! c(i,j) = y(i,j) + b(i,j) - q(i,j)*b(i+1,j)
            c(i,j) = y(i,j) - gov_spending
            tb(i,j) = y(i,j) - c(i,j)  - gov_spending
            q_default_free(i,j) = q_nodef(i_y_current, i_r_current, i_r_vol_current)
            q_default_free_new(i,j) = q_def_free_fun(i_y_current, i_r_current, i_r_vol_current)

         end if

      else !COUNTRY WAS NOT EXCLUDED IN THE PREVIOUS PERIOD

         if (default_decision(i_b_current, i_y_current, i_r_current, i_r_vol_current) > 1) then  !COUNTRY DEFAULTS
            d(i,j) = 2
            excl(i,j) = 2
            i_b_next = i_b_zero
            b(i+1,j) = b_grid(i_b_next)
            q(i,j) = q_fun(i_b_next, i_y_current, i_r_current, i_r_vol_current)  !q_paid
            y(i,j) = EXP(z_shock(i,j))  - max(zero, d0*EXP(z_shock(i,j)) + d1*EXP(z_shock(i,j))**2d+0)
            c(i,j) = y(i,j) - gov_spending
            tb(i,j) = y(i,j) - c(i,j) - gov_spending
            q_default_free(i,j) = q_nodef(i_y_current, i_r_current, i_r_vol_current)
            q_default_free_new(i,j) = q_def_free_fun(i_y_current, i_r_current, i_r_vol_current)

            qdef = zero
            WRITE(22, '(I7)') i-1 !NEED TO SUBSTRACT 1. REASON: files start saving data on period 2
            default_num = default_num + 1
          else
            d(i,j) = 1   !COUNTRY DOES NOT DEFAULT
            excl(i,j) = 1
            i_b_next = b_next_matrix(i_b_current, i_y_current, i_r_current, i_r_vol_current)
            b(i+1,j) = b_grid(i_b_next)

            q(i,j) = q_fun(i_b_next, i_y_current, i_r_current, i_r_vol_current)  !q_paid
            y(i,j) = EXP(z_shock(i,j))

            coupon = (r_prime+delta)/(1d+0 +r_prime)
            c(i,j) = y(i,j)  - gov_spending + b(i,j) * coupon - q(i,j)* (b(i+1,j) - b(i,j)*(1d+0 - delta))
            tb(i,j) = y(i,j) - c(i,j) - gov_spending
            q_default_free(i,j) = q_nodef(i_y_current, i_r_current, i_r_vol_current)
            q_default_free_new(i,j) = q_def_free_fun(i_y_current, i_r_current, i_r_vol_current)

         end if
      end if

      WRITE(21, *) y(i,j), b(i,j), q(i,j), qdef, c(i,j), tb(i,j)/y(i,j), excl(i,j), d(i,j), r(i,j), q_default_free(i,j),&
           i_y_current, i_r_current, i_r_vol_current, r_vol_matrix(i,j), &
           expected_SDF_matrix(i_y_current, i_r_current, i_r_vol_current), q_default_free_new(i,j)

   end do
end do
CLOSE(21)
end subroutine irf